#
# Licensed Materials - Property of IBM
#
# Rational ClearQuest
#
# (C) Copyright IBM Corp. 2005  All Rights Reserved
#

=head1 NAME

Diag - Command line configurable subroutine diagnostics.

=head1 SAMPLE USAGE

  use Getopt::Long;
  use Diag qw(%opt_diag $diag_stream);
  $diag_stream = \*STDOUT;               # send diag to STDOUT instead of STDERR
  GetOptions('diag|d=s%');               # set diag levels with -d pkg::subr=LVL

  sub sub1(@) {
    my $d = new Diag();                  # print entry/exit diag if LVL has '+'
    $d->output(-lvl=>1, -msg=>"L1");     # print "L1" if LVL>=1
    $d->output(-lvl=>2, -msg=>"L2");     # print "L2" if LVL>=2
  }

  sub1(qw/a b c 1 2 3/);

=head1 SAMPLE OUTPUT

  C:\>perl script.pl -d main::subr=+1
  #main::sub2# &main::sub1("a", "b", "c", "1", "2", "3")
  #main::sub1# L1

=head1 METHODS

=over 4

=cut

package Long2clob::Diag;
use strict;
use Carp;
use Data::Dumper;
use Long2clob::Globals qw($ERM_INTERNAL);

use Exporter;
our @ISA=qw/Exporter/;
our @EXPORT_OK = qw/%opt_diag $diag_stream/;
our %opt_diag;
our $diag_stream = \*STDERR;

use vars qw($VERSION);
$VERSION='1.01';

=item $dp = new Diag();

Returns an object for printing diagnostics by function name as found
in $Diag::opt_diag{package::subroutine}.

Function entry and exit messages will be printed automatically (by new
and DESTROY) if the caller diag level contains a '+' character
(i.e. "-diag Pkg::subr=1+").

=cut
sub new(%) {
    # Create a Diag object
    #
    my ($class) = shift or confess $ERM_INTERNAL;
    my $s = bless {@_}, ref($class) || $class;
    $s->{subr} = do {                   # {subr} = caller subroutine
        package DB;                     #   package DB makes caller set @args
        [caller(1)]->[3];
    };
    $s->{max} = 0;                      # {max} = max of matching diag levels
    $s->{plus} = 0;                     # {plus} = 1 if any match has a '+'
    $s->{minus} = 0;                    # {minus} = 1 if any match has a '-'
    while (my ($key,$value) = each(%opt_diag)) {
        if ($s->{subr} =~ /(?:$key)/) {
            $s->{plus} = 1 if $value =~ s/\+//g;
            $s->{minus} = 1 if $value =~ s/\-//g;
            $s->{max} = int($value) if $value > $s->{max};
        }
    }

    # Print diag for call into caller if $opt_diag{PKG::subr} contains '+'
    #
    return $s unless $s->{plus};

    # @args = printable caller args
    #
    my @args = map {
        my $arg=$_;
        # replace Pkg::Name=HASH(0xdeadbeef) with '$s' or '$class'
        if ($arg =~ /^(\w+::)*(\w+)=\w+\(\w+\)$/) {
            if (index($s->{subr}, $1) == 0) {
                $arg = '$s';
            }
            else {
                $arg = lc $1;
                $arg =~ s/^.*::/\$/;
            }
        }
        $arg; } @DB::args;
    local $"=', ';
    $s->output(-lvl=>0, -msg=>"\@_=(@args)");
    $s;
}

######################################################################

=item DESTROY

Print an exit message if diag -lvl >= 2.

=cut
sub DESTROY {
    my $s = shift;
    return unless $s->{minus};          # function exit message if LVL has '-'
    local $_ = defined($_) ? "'$_'" : "undef";
    $s->output(-lvl=>0, -msg=>"\$_=$_");
}

######################################################################

=item $ds->output(-msg=>$msg, -lvl=>$lvl)

Print a diag message if the specified level is low enough and return the number of
lines printed.

 -msg => diag message text
 -lvl => message diag level
 -pre => override default prefix (optional)

=cut
sub output(%) {
    my ($s, %p) = @_;
    ref($s) eq __PACKAGE__ or confess $ERM_INTERNAL;   # require '->' notation
    defined $p{-lvl} or confess $ERM_INTERNAL;         # require -lvl
    return if $p{-lvl} > $s->{max};                    # skip if -lvl > {max}
    defined $p{-msg} or confess $ERM_INTERNAL;         # require -msg
    $p{-pre} = "#$s->{subr}# "                         # default -pre
      unless defined $p{-pre};
    $p{-msg} =~ s/\n/\n$p{-pre}/g;                     # prefix multiline msgs
    print {$diag_stream} "$p{-pre}$p{-msg}\n";
}

1;

=back

=cut

# Local Variables: #
# mode:cperl #
# comment-column:40 #
# cperl-comment-column:40 #
# cperl-continued-brace-offset:-2 #
# cperl-indent-level:4 #
# indent-tabs-mode:nil #
# fill-column: 78 #
# End: #
