#
# Licensed Materials - Property of IBM
#
# Rational ClearQuest
#
# (C) Copyright IBM Corp. 2005,2006  All Rights Reserved
#

=head1 NAME

Long2clob::Operations - long to clob conversion operations module.

=head1 SYNOPSIS

  use Long2clob::Operation;
  my $ops = new Long2clob::Operations(pdsql=>$pdsql);
  $ops->create_table();
  @ops = $ops->get_operations();
  $ops->execute_operation($op)
  $ops->drop_table();

=head1 METHODS

=over 4

=cut

package Long2clob::Operations;
use strict;
no warnings 'once';
use Carp;
use Data::Dumper;
use Long2clob::PDSQL qw/$COLHEAD/;
use Long2clob::Diag;
use Long2clob::Globals qw(/\$ERM_/ /\%Fixed_/ $LONG2CLOB);
use Exporter;
our @ISA = qw/Exporter/;
our @EXPORT_OK = qw(timestamp);

use vars qw($VERSION);
$VERSION='1.10';

# Define the possible states the database transitions through during
# conversion.
#
push @EXPORT_OK, qw($DBS_LONG $DBS_LOCKED $DBS_INITIALIZED $DBS_CONVERTED
                    $DBS_CONVERTING $DBS_FINALIZED $DBS_UNLOCKED $DBS_CLOB);
our $DBS_LONG = "LONG";
our $DBS_LOCKED = "LOCKED";
our $DBS_INITIALIZED = "INITIALIZED";
our $DBS_CONVERTING = "CONVERTING";
our $DBS_CONVERTED = "CONVERTED";
our $DBS_FINALIZED = "FINALIZED";
our $DBS_UNLOCKED = "UNLOCKED";
our $DBS_CLOB = "CLOB";

# Define operation names, export them, and map them to subroutines
#
our %OP_SUBS;
our ($OPN_INITIALIZE, $OPN_FINALIZE,
     $OPN_INDEX_CREATE, $OPN_INDEX_DROP, $OPN_INDEX_RENAME,
     $OPN_TABLE_ADD_COL, $OPN_TABLE_CREATE, $OPN_TABLE_DROP,
     $OPN_TABLE_INSERT, $OPN_TABLE_RENAME, $OPN_TABLE_SET_COL);

for (qw(INITIALIZE FINALIZE
        CONNECT_OPTS INDEX_CREATE INDEX_DROP INDEX_RENAME TABLE_ADD_COL
        TABLE_CREATE TABLE_DROP TABLE_INSERT TABLE_RENAME TABLE_SET_COL))
{
    eval "\$OPN_$_ = '$_';";            # $OPN_XXXX = 'XXXX';
    push(@EXPORT_OK, "\$OPN_$_");       # @EXPORT_OK = qw($OPN_XXXX);
    $OP_SUBS{$_}= \&{'_op_'.lc($_)};    # $OP_SUBS{XXXX} = \&_op_xxxx;
}

# Operation STATUSES
#
push @EXPORT_OK, qw($OPS_COMPLETED $OPS_REQUIRED $OPS_DEFERRED);
our $OPS_COMPLETED = 'COMPLETED';       # operation successfully completed
our $OPS_REQUIRED = 'REQUIRED';         # this operation must succeed
our $OPS_DEFERRED = 'DEFERRED';         # this operation has been deferred

# OP_TABLE schema constants
our $OP_TABLE = "RATL_LONGTOCLOB_OPERATIONS";
our $OP_SEQUENCE = "RATL_LONGTOCLOB_SEQUENCE";
our ($OP_NAME, $OP_ARGS, $OP_STATUS, $OP_SEQID) = qw(op_name op_args op_status op_seqid);
our ($OP_NAME_LEN, $OP_ARGS_LEN, $OP_STATUS_LEN) = (50, 200, 50);

# Actions on operations
push @EXPORT_OK, qw($OPA_DEFER $OPA_EXECUTE $OPA_SKIP);
our $OPA_DEFER = "DEFER";
our $OPA_EXECUTE = "EXECUTE";
our $OPA_SKIP = "SKIP";

######################################################################

=item my $ts = timestamp;

=item my $ts = timestamp($time);

=item my $ts = timestamp(@localtime);

Return the current time as a UTC string without the trailing Z for GM time,
i.e. "yyyy-mm-ddThh:mm:ss".  By default, the time is based on the array
returned by localtime(time()), but the caller can specify the time to use via
a scalar as returned by time(), or via an array returned by localtime(time()).

=cut
sub timestamp(@) {
    @_ = (time) if $#_ == -1;
    @_ = (localtime($_[0])) if $#_ == 0;
    $#_ == 8 or confess $ERM_INTERNAL;
    my ($sec,$min,$hr,$mday,$mon,$year,$wday,$yday,$isdst) = @_;
    sprintf("%4d-%02d-%02dT%02d:%02d:4:%02d",
            $year+1900,$mon+1,$mday,$hr,$min,$sec);
}

######################################################################

=item $ops = new Long2clob::Operations(pdsql=>$pdsql);

Return an object for handling long to clob conversion operations.

=cut
sub new(%) {
    my $d = new Long2clob::Diag();
    my ($class) = shift or confess $ERM_INTERNAL;
    my $s = bless {@_}, ref($class) || $class;
    ref($s->{pdsql}) eq 'Long2clob::PDSQL' or confess $ERM_INTERNAL;
    $s;
}

######################################################################

=item my $status = $ops->get_conversion_status()

Return one of the following states:

 $DBS_LONG        = unlocked, LONG
 $DBS_LOCKED      = locked, no ops table
 $DBS_INITIALIZED = locked, w/ops, none completed
 $DBS_CONVERTING  = locked, w/ops, w/completed, w/required
 $DBS_CONVERTED   = locked, w/ops, w/completed, none required, no SET_CONNECT_OPTS
 $DBS_FINALIZED   = locked, w/ops, w/completed, none required, w/SET_CONNECT_OPTS
 $DBS_UNLOCKED    = unlocked, CLOB, w/ops
 $DBS_CLOB        = unlocked, CLOB, no ops table

=cut
sub get_conversion_status() {
    my $ds = new Long2clob::Diag();
    my ($s) = @_;
    my $long = $s->{pdsql}->is_long_database();
    my $locked = $s->{pdsql}->is_locked_for_reason($LONG2CLOB);
    my @ops = eval { $s->get_operations() };
    my @required = grep { $_->{status} eq $OPS_REQUIRED } @ops;
    my $required = 1 if $#required > -1;
    my @completed = grep { ($_->{status} eq $OPS_COMPLETED and
                            $_->{name} ne $OPN_FINALIZE and
                            $_->{name} ne $OPN_INITIALIZE) } @ops;
    my $completed = 1 if $#completed > -1;
    my $has_finalize = 1 if grep { $_->{name} eq $OPN_FINALIZE } @ops;
    my $has_initialize = 1 if grep { $_->{name} eq $OPN_INITIALIZE } @ops;

    return ($DBS_LONG, "See the 'lock' subcommand to lock the database.")
      if !$locked && $long;
    return ($DBS_LOCKED, "See the 'initialize' subcommand to plan conversion operations.")
      if $locked && !$has_initialize;
    return ($DBS_INITIALIZED, "See the 'convert' subcommand to convert the database.")
      if $locked && $has_initialize && !$completed;
    return ($DBS_CONVERTING, "See the 'convert' subcommand to resume conversion.")
      if $locked && $has_initialize && $completed && $required;
    return ($DBS_CONVERTED, "See the 'finalize' subcommand to update its connect options.")
      if $locked && $has_initialize && $completed && !$required && !$has_finalize;
    return ($DBS_FINALIZED, "See the 'unlock' subcommand to reenable access.")
      if $locked && $has_initialize && $completed && !$required && $has_finalize;
    return ($DBS_UNLOCKED, "See the 'cleanup' subcommand to execute deferred operations.")
      if !$locked && !$long && $has_initialize;
    return ($DBS_CLOB, "This is a CLOB database and needs no conversion.")
      if !$locked && !$long && !$has_initialize;
    confess "$ERM_INTERNAL:  The database is in an unknown state and should be restored from backup.\n";
}

######################################################################

=item $ops->drop_table()

Drop the OP_TABLE table.

=cut
sub drop_table() {
    my $ds = new Long2clob::Diag();
    ref($_[0]) eq __PACKAGE__ or confess $ERM_INTERNAL;
    my ($s) = @_;
    $s->{pdsql}->drop_table($OP_TABLE);
    $s->{pdsql}->drop_sequence($OP_SEQUENCE);
}

######################################################################

=item $ops->create_table()

Create the OP_TABLE table and the OP_TABLE sequence.

=cut
sub create_table() {
    my $ds = new Long2clob::Diag();
    ref($_[0]) eq __PACKAGE__ or confess $ERM_INTERNAL;
    my ($s) = @_;
    $s->{pdsql}->pdsql(<<EOF);
create table $OP_TABLE(
  $OP_SEQID number primary key,
  $OP_NAME varchar2($OP_NAME_LEN) not null,
  $OP_ARGS varchar2($OP_ARGS_LEN) not null,
  $OP_STATUS varchar2($OP_STATUS_LEN) not null);
EOF
    $s->{pdsql}->pdsql(<<EOF);
create sequence $OP_SEQUENCE increment by 1 start with 1;
EOF
}

######################################################################

=item my @ops = $ops->get_operations()

Return array with all op hash refs in table.

=cut
sub get_operations() {
    my $ds = new Long2clob::Diag();
    ref($_[0]) eq __PACKAGE__ or confess $ERM_INTERNAL;
    my ($s,$status) = @_;
    my $sql = <<EOF;
select $OP_NAME $OP_NAME$COLHEAD, $OP_ARGS $OP_ARGS$COLHEAD, $OP_STATUS $OP_STATUS$COLHEAD, $OP_SEQID $OP_SEQID$COLHEAD
  from $OP_TABLE order by $OP_SEQID;
EOF
    my @ops = $s->{pdsql}->pdsql_select($sql, 4);
    # convert arrays to hashes
    @ops = map {{ name => $_->[0], args=>$_->[1], status=>$_->[2], seqid=>$_->[3]}} @ops;
}

######################################################################

=item $ops->plan_operation($op)

Add a new operation to OP_TABLE.  Returns the added $op.

=cut
sub plan_operation(%) {
    my $ds = new Long2clob::Diag();
    ref($_[0]) eq __PACKAGE__ or confess $ERM_INTERNAL;
    my ($s, %op) = @_;
    confess "$ERM_INTERNAL: missing parameter: {name}" unless exists $op{name};
    confess "$ERM_INTERNAL: missing parameter: {args}" unless exists $op{args};
    $op{status} = $OPS_REQUIRED unless exists $op{status};
    printf("Planning:  %-10s %-15s %-15s\n", $op{status}, $op{name}, $op{args})
      unless $op{silent};
    $s->{pdsql}->pdsql(<<EOF);
        insert into $OP_TABLE($OP_NAME, $OP_ARGS, $OP_STATUS, $OP_SEQID)
        values('$op{name}', '$op{args}', '$op{status}', $OP_SEQUENCE.nextval);
EOF
    return \%op;
}

######################################################################

=item $ops->execute_operation($op);

This is a front end controller/delegate to execute an operation.  Only
successes are updated in the OP_TABLE.op_status field.

=cut
sub execute_operation(\%) {
    my $ds = new Long2clob::Diag();
    @_==2 or confess $ERM_INTERNAL;
    my ($s, $op) = @_;
    #
    # Get the SQL for this operation and remove excess whitespace.
    #
    my $sql = $s->get_sql_statement($op);
    print("  $sql\n\n") if $::opt_verbose;
    $sql =~ s/\n//g;
    $sql =~ s/\s+/ /g;
    $sql =~ s/^ //;
    $sql =~ s/ $//;
    $ds->output(-lvl=>1, -msg=>Dumper($op));
    #
    # Execute SQL inside this block with local signal handlers.
    #
    {
        # delay SIGINT/SIGTERM handling until after PDSQL has returned
        my @signals;
        local $SIG{'INT'} = sub { push @signals, 'SIGINT(2)'; };
        local $SIG{'TERM'} = sub { push @signals, 'SIGTERM(15)'; };

        eval {
            $s->{pdsql}->pdsql($sql) if ! $::opt_noop;
        };
        if ($@) {
            $s->handle_error($op);
        }
        $s->set_operation_status($op, $OPS_COMPLETED) if ! $::opt_noop;
        if ($#signals >= 0) {
            # die "Terminating on signal: @signals\n";
            exit(3);
        }
    }
    return 1;
}

######################################################################

=item $ops->set_operation_status($op)

Update the $op in OP_TABLE.

=cut
sub set_operation_status(\%;$) {
    my $ds = new Long2clob::Diag();
    my ($s, $op, $status) = @_;
    $op->{status} = $status if defined $status;
    my $sql = <<EOF;
update $OP_TABLE op set $OP_STATUS = '$op->{status}' where op.$OP_NAME = '$op->{name}' AND op.$OP_ARGS = '$op->{args}';
EOF
    $s->{pdsql}->pdsql($sql) if ! $::opt_noop;
}

######################################################################

=item my $sql = $ops->get_sql_statement($op);

Return the SQL statement for an operation.  Dies if the operation or
argument is invalid.

=cut
sub get_sql_statement($) {
    my $ds = new Long2clob::Diag();
    @_ == 2 or confess $ERM_INTERNAL;
    my ($s, $op) = @_;
    my $subr = $OP_SUBS{$op->{name}};
    defined($subr) or confess "$ERM_OPERATIONS: invalid operation name: '$op->{name}'";
    my $sql = &$subr($s,$op);
    defined($sql) or confess "$ERM_OPERATIONS: invalid operation: '$op->{name}'  '$op->{args}'";
    chomp $sql;
    $_ = $sql;
}

######################################################################

=item $ops->handle_error($op);

Handle error in $@ from executing pdsql for $op, then die or return.

=cut
sub handle_error($) {
    my $sqlerr = $@;
    $@ = '';
    my $ds = new Long2clob::Diag();
    @_ == 2 or confess $ERM_INTERNAL;
    my ($s, $op) = @_;
    local $_ = $op->{name};
    if (! $sqlerr) {
        return;                         # no error occurred
    }
    if ($::opt_fatal) {
        die $sqlerr;                    # always die with -fatal
    }
    if ($_ eq $OPN_INDEX_CREATE) {
        return if $sqlerr=~/ORA-00955/; # index exists (ok)
        die $sqlerr;
    }
    if ($_ eq $OPN_INDEX_RENAME) {
        return if $sqlerr=~/ORA-01418/; # index missing (ok)
        die $sqlerr;
    }
    if ($_ eq $OPN_INDEX_DROP) {
        return if $sqlerr=~/ORA-01418/; # index missing (ok)
        die $sqlerr;
    }
    if ($_ eq $OPN_TABLE_DROP) {
        return if $sqlerr=~/ORA-00942/; # table missing (ok)
        die $sqlerr;
    }
    if ($_ eq $OPN_TABLE_RENAME) {
        return if $sqlerr=~/ORA-00942/; # table missing (ok)
        die $sqlerr;
    }
    if ($_ eq $OPN_TABLE_ADD_COL) {
        die $sqlerr;                    # never ok
    }
    if ($_ eq $OPN_TABLE_CREATE) {
        return if $sqlerr=~/ORA-00955/; # table exists (ok)
        die $sqlerr;
    }
    if ($_ eq $OPN_TABLE_INSERT) {
        die $sqlerr;                    # never ok
    }
    if ($_ eq $OPN_TABLE_SET_COL) {
        die $sqlerr;                    # never ok
    }
    die "handle_error: unknown operation: $_\n$sqlerr";
}


######################################################################
# All _op_operation_name methods are called by name from 'get_sql_statement'.
######################################################################

sub _op_index_create(\%) {
    my $ds = new Long2clob::Diag();
    my ($s, $op) = @_;
    #
    # Map index_name to index_create then look up the "create index" statement.
    #
    my $desc = $s->{pdsql}->is_schema_repo() ? \%Fixed_SR_Info : \%Fixed_UserDB_Info;
    my @names = map { ref($_)eq'ARRAY' ? @$_ : $_ } @{$desc->{index_name}};
    my @creates = map { ref($_)eq'ARRAY' ? @$_ : $_ } @{$desc->{index_create}};
    my %create = map { $names[$_] => $creates[$_] } 0..$#names;
    $_ = $create{$op->{args}};
}

sub _op_index_rename(\%) {
    my $ds = new Long2clob::Diag();
    my ($s, $op) = @_;
    my @args = split(/,/, $op->{args});
    $_ = "alter index $args[0] rename to $args[1];";
}

sub _op_index_drop(\%) {
    my $ds = new Long2clob::Diag();
    my ($s, $op) = @_;
    $_ = "drop index $op->{args};";
}

sub _op_table_add_col(\%) {
    my $ds = new Long2clob::Diag();
    my ($s, $op) = @_;
    my @args = split(/,/, $op->{args});
    $_ = "alter table $args[0] ADD $args[1] CLOB;";
}

sub _op_table_create(\%) {
    my $ds = new Long2clob::Diag();
    my ($s, $op) = @_;
    my $table_name = $op->{args};
    #
    # Map table_name to column_create then look up column_create SQL.
    #
    my $desc = $s->{pdsql}->is_schema_repo() ? \%Fixed_SR_Info : \%Fixed_UserDB_Info;
    my %column_create = map {
        $desc->{table_name}->[$_] => $desc->{column_create}->[$_]
    } 0..$#{$desc->{table_name}};
    my $column_create = $column_create{$table_name} or return undef $_;
    $_ = "create table ${table_name}($column_create);";
}

sub _op_table_drop(\%) {
    my $ds = new Long2clob::Diag();
    my ($s, $op) = @_;
    my @args = split(/,/, $op->{args});
    my $table = $args[0];
    if ($#args == 1 && $args[1] =~ /\/rows/i) {
        # If args are "TABLE_NAME,/rows" then just delete the rows.
        $_ = "delete from $table;";
    }
    else {
        $_ = "drop table $table cascade constraints;";
    }
}

sub _op_table_insert(\%) {
    my $ds = new Long2clob::Diag();
    my ($s, $op) = @_;
    my $table_name = $op->{args};
    #
    # Map table_name to select_clause then look up select_clause for table.
    #
    my $desc = $s->{pdsql}->is_schema_repo() ? \%Fixed_SR_Info : \%Fixed_UserDB_Info;
    my %select_clause = map {
        $desc->{table_name}->[$_] => $desc->{select_clause}->[$_]
    } 0..$#{$desc->{table_name}};
    my $select_clause = $select_clause{$table_name} or return undef $_;
    $_ = "insert INTO ${table_name}(SELECT $select_clause FROM ${table_name}_LONG);";
}

sub _op_table_rename(\%) {
    my $ds = new Long2clob::Diag();
    my ($s, $op) = @_;
    my @args = split(/,/, $op->{args});
    $_ = "alter table $args[0] rename to $args[1];";
}

sub _op_table_set_col(\%) {
    my $ds = new Long2clob::Diag();
    my ($s, $op) = @_;
    my @args = split(/,/, $op->{args});
    @args == 3 or return undef $_;
    my ($userentity_db, $fielddef_db, $fielddef_id) = @args;
    $_ = <<EOF;
update $userentity_db e
  set $fielddef_db =
    (SELECT data
     FROM multiline_text mt
     WHERE mt.entity_dbid = e.dbid AND mt.entity_fielddef_id = $fielddef_id);
EOF
}

1;

=back

=cut

# Local Variables: #
# mode:cperl #
# comment-column:40 #
# cperl-comment-column:40 #
# cperl-continued-brace-offset:-2 #
# cperl-indent-level:4 #
# indent-tabs-mode:nil #
# fill-column: 78 #
# End: #
