#
# Licensed Materials - Property of IBM
#
# Rational ClearQuest
#
# (C) Copyright IBM Corp. 2005,2006  All Rights Reserved
#

package Long2clob::PDSQL;
use strict;
use Carp;
use Data::Dumper;
use File::Temp qw/tempfile tempdir/;
use Win32::TieRegistry;

use Long2clob::Diag;
use Long2clob::Globals qw(/\$ERM_/);

use vars qw($VERSION);
$VERSION='1.09';

######################################################################
# Variables
######################################################################

use Exporter;
our @ISA = qw/Exporter/;
our @EXPORT_OK = qw/$COLHEAD $SR2 $SR3/;

our $COLHEAD  = "_COLHEAD";             # Required col-head suffix for selects
our $VERSION_SR2 = 'v2003.06.12';       # Version string for SR2
our $VERSION_SR3 = 'v2003.06.13+';      # Version string for SR3+

# Registry key under which DBSETS are stored.
#
our $DBSETS_KEY = "HKEY_CURRENT_USER\\Software\\Rational Software\\ClearQuest\\2003.06.00\\Core\\Databases\\";

######################################################################

=item $exe_version = Long2clob::PDSQL::exe_version();

Return the rough version of PDSQL.EXE found.  For ClearQuest GA through SR2,
it will return "v2003.06.12".  For ClearQuest SR3 through SR5, it will return
"v2003.06.13+".  For other versions, i.e. version 7, it will return ''.

=cut
our $exe_version = undef;               # only determine exe_version once
sub exe_version
{
    my $ds = new Long2clob::Diag();
    return $exe_version if defined $exe_version;
    # read the help output of pdsql.exe
    local $/;
    open(PDSQL, "pdsql |") or confess "can't fork: $!";
    my $text = <PDSQL>;
    close(PDSQL);
    # look for identifying strings in the help output
    my $SR2_TEXT = 'LOB_TYPE=<LONG,CLOB> default is LONG';
    my $SR3_TEXT = 'LOB_TYPE=<LONG,CLOB> default is CLOB';
    if (index($text, $SR2_TEXT) >= 0) {
        $exe_version = $VERSION_SR2;
    }
    elsif (index($text, $SR3_TEXT) >= 0) {
        $exe_version = $VERSION_SR3;
    }
    else {
        $exe_version = '';
    }
    return $exe_version;
}

######################################################################

=item $db = new Long2clob::PDSQL(%DBOPTS);

Return a PDSQL object for handling PDSQL statements.  %DBOPTS has the
following values which are passed on as command line parameters to PDSQL.EXE:

  u   => -u <userid>
  p   => -p <password>
  v   => -v <ora7 or ora8> with SR2
  co  => -co <connect_options>
  s   => -s <sr3-server or sr2-alias>
  db  => -db <sid>

=cut
sub new {
    my $ds = new Long2clob::Diag();
    my ($class) = shift or confess $ERM_INTERNAL;
    my %p = (@_);
    my $s = bless {}, ref($class) || $class;

    # Look for parameters required by PDSQL
    #
    my @missing;
    push(@missing, '-u <user>') unless (defined $p{'u'});
    push(@missing, '-p <password>') unless (defined $p{'p'});
    # check SR3 PDSQL parameters
    #
    if (exe_version() eq $VERSION_SR3)
    {
        push(@missing, '-s <host>') unless (defined $p{'s'});
        push(@missing, '-db <sid>') unless (defined $p{'db'});
        $p{'v'} = 'ora' unless (defined $p{'v'});
    }
    elsif (exe_version() eq $VERSION_SR2)       # check SR2 PDSQL parameters
    {
        push(@missing, '-s <alias>') unless (defined $p{'s'});
        push(@missing, '-v <ora7 or ora8>') unless (defined $p{'v'});
        push(@missing, '-co HOST=<host>;SID=<sid>;SERVER_VER=<version>;CLIENT_VER=<version>')
            unless (defined $p{'co'});
    }
    else                                        # invalid PDSQL version found
    {
        die "$ERM_PDSQL:  PDSQL.EXE is not v2003.06.12 or later.\n";
    }
    # report missing parameters
    if (@missing > 0) {
        die "$ERM_USAGE:  missing database connection information:\n    "
          . join("\n    ", @missing)
            . "\n";
    }
    # Set up the PDSQL.EXE command line with arguments
    $s->{'cmdline'} = "pdsql";
    for (qw/u p co s db dsn v/) {
        $s->{'cmdline'} .= " -$_ \"$p{$_}\"" if defined $p{$_};
    }
    $s->{'cmdline'} .= " -errors_are_fatal -noprompt";
    $s->{'args'} = \%p;
    # Set up a {tmp} directory that will be cleaned up on exit
    if (! exists $s->{tmp}) {
        (my $pkg = lc(__PACKAGE__)) =~ s/::/_/;
        $s->{tmp} = tempdir("$pkg-XXXXX", TMPDIR=>1, CLEANUP=>1);
    }
    return $s;
}

######################################################################

=item $db->get_connect_options($site_name)

Get the connect options for a site (user database) out of the master database
MASTER_DBS table.

=cut
sub get_connect_options {
    my $ds = new Long2clob::Diag();
    @_==2 && ref($_[0]) eq __PACKAGE__ or confess "$ERM_INTERNAL";
    my ($s, $site_name) = @_;
    my $sql =<<EOF;
select connect_options connect_options$COLHEAD
  from MASTER_DBS
  where name = '$site_name';
EOF
    my @rows = $s->pdsql_select($sql, 1);
    $#rows == 0 or confess "$ERM_INTERNAL\nToo many databases matched select:\n  $sql\n";
    return $rows[0]->[0];
}

######################################################################

=item $db->set_connect_options($site_name, $new_co)

Set the connect options for a site (user database) in the master database
MASTER_DBS table.

=cut
sub set_connect_options {
    my $ds = new Long2clob::Diag();
    @_==3 && ref($_[0]) eq __PACKAGE__ or confess "$ERM_INTERNAL";
    my ($s, $site_name, $new_co) = @_;
    my $sql =<<EOF;
update MASTER_DBS
  set connect_options = '$new_co'
  where name = '$site_name';
EOF
    my $count = $s->pdsql_update($sql);
    confess "$ERM_INTERNAL\nUpdated $count connect_options entries (expected 1)!"
      unless $count == 1;
}

######################################################################

=item if ($db->is_reachable()) ...

Return true if the database is reachable.

=cut
sub is_reachable() {
    my $ds = new Long2clob::Diag();
    @_==1 or confess $ERM_INTERNAL;
    my ($s) = @_;
    eval { $s->pdsql("quit"); };
    return $@ ? 0 : 1;
}

######################################################################

=item @user_entities = $pdsql->getAllUserEntities()

Return rows of [entity, fielddef, fielddef_id]

=cut
sub getAllUserEntities() {
    my $ds = new Long2clob::Diag();
    @_==1 or confess $ERM_INTERNAL;
    my ($s) = @_;
    my $tablename = "RETRIEVALONLY";
    #
    # The SQL below looks up user defined entitydefs and fielddefs..  The
    # exclusion of family entitydefs is because they are like abstract types
    # have no tables associated.  The select to identify a 'family' comes from
    # cqmetadata/addatabase.cpp ... literally "NOT IN(1, 3, 5, 7)".
    #
    my $sql = <<EOF;
select DISTINCT
    TRIM(e.db_name) i_entity_db_name$COLHEAD,
    TRIM(f.db_name) i_fielddef_db_name$COLHEAD,
    f.id i_fielddef_id$COLHEAD
  FROM fielddef f, entitydef e
  WHERE f.type = 2 AND e.id = f.entitydef_id
  AND e.is_family NOT IN(1,3,5,7)
  ORDER BY i_entity_db_name$COLHEAD, i_fielddef_db_name$COLHEAD;
EOF
    my @rows = $s->pdsql_select($sql, 3);
    confess "$ERM_PDSQL: Zero user entity rows retrieved (highly unlikely that is correct)!\n$_\n"
      if @rows == 0;
    return @rows;
}

######################################################################

=item @lines = $db->pdsql($sql);

Pass $sql to the 'pdsql' command.  Store output in $_ and return an array of
lines.  Die with error message and full output if pdsql encounters an error.

=cut
sub pdsql($) {
    my $ds = new Long2clob::Diag();
    @_==2 or confess $ERM_INTERNAL;
    my ($s, $sql) = @_;
    $ds->output(-lvl=>1, -msg=>"CMD=$s->{cmdline}");
    $sql =~ s/[\n\s]+/ /g;              # reduce NL/WS to single space
    $sql =~ s/[ ;]*$/;/;                # make sure $sql ends with ';'
    confess "$ERM_PDSQL: PDSQL commands must be lower case\nSQL:\n$sql\n"
      if $sql =~ /^[A-Z]/;
    $ds->output(-lvl=>1, -msg=>"SQL=$sql");

    # Perl equivalent of "echo $sql | pdsql.exe >& $tempfile"
    my $tempfile;
    (undef, $tempfile) = tempfile('pdsql-XXXXX', DIR=>$s->{tmp}, SUFFIX=>'.out');
    $ds->output(-lvl=>2, -msg=>"TMP=$tempfile");
    open(PDSQL, "| $s->{cmdline} > $tempfile 2>&1")
      or confess "can't fork: $!";
    print PDSQL "longmax 500;\n";       # prevent PDSQL truncations
    print PDSQL $sql;
    close(PDSQL);

    # Read @lines from $tempfile, stripping leading and trailing whitespace
    my @lines;
    my $raw_output = '';
    open(OUTPUT, "< $tempfile") or confess "can't open $tempfile: $!";
    while (<OUTPUT>) {
        $raw_output .= $_;      # save raw output for debugging
        chomp;                  # discard newline
        s/^(\>\s*)*//;          # remove leading '>'
        s/\s+/ /g;              # convert WS to a single spaces
        s/^ //;                 # remove leading space
        s/ $//;                 # remove trailing space
        next if $_ eq '';       # discard blank lines
        push @lines, $_;
    }
    close(OUTPUT);
    unlink $tempfile;

    # check for errors and return
    confess "$ERM_ORACLE: $1\n$s->{cmdline}\n$sql\n$raw_output\n"
      if $raw_output =~ /(ORA-\d+):/s;
    confess "$ERM_PDSQL: $s->{cmdline}\n$sql\n$raw_output\n"
      if $raw_output =~ /^EXCEPTION:/m;
    $_=join("\n", @lines);
    return wantarray ? @lines : $_;
}

######################################################################

=item @rows = $db->pdsql_select($sql, $expected)

Execute a PDSQL select statement and return the results in an ARRAY of
rows, where each row is an ARRAY of columns.

 $sql:      the PDSQL statement to execute.
 $expected: the number of columns expected back.
 @rows:     $rows[$rownum]->[$colnum] results.

Note: Only the last column can have embedded whitespace.

=cut
sub pdsql_select($$) {                  # PDSQL SELECT ($expect cols per row)
    my $ds = new Long2clob::Diag();
    @_==3 or confess $ERM_INTERNAL;
    my ($s, $sql, $expected) = @_;
    my @rows = $s->pdsql($sql);         # execute the select statement
    my $raw_out = $_;
    my $hdr = shift @rows;              # remove the header row
    my $nc = () = $hdr =~ /$COLHEAD/g;  # use it to verify $expected columns
    confess "$ERM_PDSQL_SELECT: expected $expected columns, got $nc\nSTATEMENT:\n$sql\nRESULT:\n$raw_out\n"
      unless $nc == $expected;
    @rows = map {                       # split each row into array of columns
        [ split(/\s+/, $_, $expected) ]
    } @rows;
    return @rows;
}

######################################################################

=item $nrows = $db->pdsql_update($sql)

=item $nrows = $db->pdsql_insert($sql)

=item $nrows = $db->pdsql_delete($sql)

Execute an $sql update, insert, or delete command.  Return the number
of rows affected.

=cut
sub pdsql_update($);
no warnings qw/once/;
*pdsql_delete = *pdsql_insert = \&pdsql_update;
sub pdsql_update($) {
    my $ds = new Long2clob::Diag();
    @_==2 or confess $ERM_INTERNAL;
    my ($s, $sql) = @_;
    my @lines = $s->pdsql($sql);
    if ($lines[0] =~ /^(\d+) rows affected.$/) {
        return $1;
    } else {
        return 0;
    }
}

######################################################################
# the following are a bunch of subs which bundle common error handling
######################################################################

######################################################################

sub drop_sequence($) {
    my $ds = new Long2clob::Diag();
    @_==2 or confess $ERM_INTERNAL;
    my ($s, $name) = @_;
    eval '$s->pdsql("drop sequence $name;")';
    $@='' if $@ =~ /ORA-0*2289/;       # ORA-2289 is fine (seq not found)
    die $@ if $@;
    return 1;
}

######################################################################

sub drop_table($) {
    my $ds = new Long2clob::Diag();
    @_==2 or confess $ERM_INTERNAL;
    my ($s, $tablename) = @_;
    eval '$s->pdsql("drop table $tablename cascade constraints;")';
    $@='' if $@ =~ /ORA-0*942/;   # ORA-942 is fine (table not found)
    die $@ if $@;
    return 1;
}

######################################################################

sub is_schema_repo() {          # return true if db is a schema repo
    my $ds = new Long2clob::Diag();
    @_==1 or confess $ERM_INTERNAL;
    my ($s) = @_;
    if (! defined $s->{is_schema_repo}) {
        my @rows = $s->pdsql_select("select count(*) DaCount$COLHEAD from user_tables where table_name = 'MASTER_DBS';", 1);
        if ($rows[0][0] != 1) {
            $s->{is_schema_repo} = 0;
        } else {
            $s->{is_schema_repo} = 1;
        }
    }
    return $s->{is_schema_repo};
}

######################################################################

sub get_lock_table_name() {        # lock table name is different for the SR
    my $ds = new Long2clob::Diag();
    @_==1 or confess $ERM_INTERNAL;
    my ($s) = @_;
    if ($s->is_schema_repo()) {
        return "master_global";
    } else {
        return "dbglobal";
    }
}

######################################################################

sub is_long_database() {           # return true for LONG databases (vs CLOB)
    my $ds = new Long2clob::Diag();
    @_==1 or confess $ERM_INTERNAL;
    my ($s) = @_;
    if ($s->is_schema_repo()) {
        # look in SR DB for LONG column in the MASTER_GLOBAL table
        if ($s->pdsql("columns MASTER_GLOBAL;") =~ /\bROLL_USER_IDS_SCRIPT\s+LONG\b/) {
            return 1;
        } else {
            return 0;
        }
    } else {
        # look in USER DB for LONG column in the USER_BLOB table
        if ($s->pdsql("columns USER_BLOB;") =~ /\bDATA\s+LONG\b/) {
            return 1;
        } else {
            return 0;
        }
    }
    return 1;
}

######################################################################

=item $db->lock_for_reason($reason)

Lock the database for the specified reason.  Die if it is already locked.

=cut
sub lock_for_reason($) {
    my $ds = new Long2clob::Diag();
    @_==2 or confess $ERM_INTERNAL;
    my ($s, $reason) = @_;
    if ($s->is_locked_for_reason($reason)) {
        die "Lock error: database already locked for $reason\n";
    }
    my $locktable = $s->get_lock_table_name();
    my $sql = "update $locktable set db_locked = 1, db_locked_info = '$reason' where db_locked = 0;";
    my $count = $s->pdsql_update($sql);
    if ($count == 0) {
        confess <<EOF;
$ERM_LOCK_FAILED
Locking database affected $count rows - another lock may exist.
EOF
    }
    return 1;
}

######################################################################

=item $db->unlock_for_reason($reason)

Unlock a database that is locked for $reason.

=cut
sub unlock_for_reason($) {
    my $ds = new Long2clob::Diag();
    @_==2 or confess $ERM_INTERNAL;
    my ($s, $reason) = @_;
    if (! $s->is_locked_for_reason($reason)) {
        die "Unlock error: database is not locked for $reason";
    }
    my $lock_table = $s->get_lock_table_name();
    my $sql = "update $lock_table set db_locked = 0, db_locked_info = '' where db_locked = 1;";
    my $count = $s->pdsql_update($sql);
    if ($count != 1) {
        confess "Unlock error: $count rows affected (database unlocked or locked for another reason?)\n";
    }
    return 1;
}

######################################################################

=item $db->is_locked_for_reason($reason)

Return true if db is locked for the specified reason.

=cut
sub is_locked_for_reason($) {
    my $ds = new Long2clob::Diag();
    @_==2 or confess $ERM_INTERNAL;
    my ($s, $reason) = @_;

    # return true if db_locked ==1 AND db_locked_info = "'$reason, false otherwise.
    my $tablename = $s->get_lock_table_name();
    my $sql = "select count(*) DaCount$COLHEAD from $tablename where db_locked = 1 and db_locked_info = '$reason';";
    my @rows = $s->pdsql_select($sql, 1);
    if ($rows[0][0] == 1) {
        return 1;
    } else {
        return 0;
    }
}

######################################################################

=item my @locks = $db->is_locked()

Return the reason(s) the database is locked, or an empty list if it is not locked.

=cut
sub is_locked() {
    my $ds = new Long2clob::Diag();
    @_==1 or confess $ERM_INTERNAL;
    my ($s, $reason) = @_;
    my $table = $s->get_lock_table_name();
    my $sql = "select db_locked_info db_locked_info$COLHEAD from $table where db_locked = 1;";
    my @rows = $s->pdsql_select($sql, 1);
    #
    # Just one column so translate 'rows of columns' into just 'rows'
    my @locks = map { $_->[0] } @rows;
    return @locks;
}

######################################################################

=item $db->update_registry()

=cut
sub update_registry {
    my $ds = new Long2clob::Diag();
    my ($s) = @_;

    # The connect options must be updated from a new process, otherwise the
    # stale database connect options are kept.
    #
    my ($fh, $tempfile) = tempfile('pdsql-XXXXX', DIR=>$s->{tmp}, SUFFIX=>'.pl');

    # Create a script to update the connect options.  It takes one parameter,
    # the dbset name to refresh.  It returns 0 for no access error, and 1 for
    # access error
    #
    print {$fh} <<'EOF';
require 'CQPerlExt.pm';
my $s = CQSession::Build();
eval { $s->GetAccessibleDatabases("MASTR", "", $ARGV[0]); };
my $err = $@;
CQSession::Unbuild($s);
if ($err) {
  print STDERR $err;
  exit 1;
}
else {
  exit 0;
}
EOF
    close($fh);

    # Update the connect options for every dbset in the registry
    #
    my $encountered_locked_mastr = 0;
    my @dbsets = grep { s/\\$// } keys %{$Registry->{$DBSETS_KEY}};
    for my $dbset (@dbsets) {
        my $status = system("cqperl $tempfile $dbset 2>NUL");
        if ($status != 0) {
            $encountered_locked_mastr = 1;
        }
    }
    unlink($tempfile);
    if ($encountered_locked_mastr) {
        return 0;
    } else {
        return 1;
    }
}

######################################################################

=item @names = $db->getUserEntityTableNames()

Return list of all all user entity table names.

=cut
sub getUserEntityTableNames(@) {
    my $ds = new Long2clob::Diag();
    my ($s) = @_;
    my $tablename = "RETRIEVALONLY";
    my $sql = <<HERE;
select DISTINCT TRIM(e.db_name) i_entity_db_name$COLHEAD,
                TRIM(f.db_name) i_fielddef_db_name$COLHEAD,
                f.id i_fielddef_id$COLHEAD
  FROM fielddef f, entitydef e
  WHERE f.type = 2 AND e.id = f.entitydef_id
  order by i_entity_db_name$COLHEAD, i_fielddef_db_name$COLHEAD;
HERE
    my @rows = $s->pdsql_select($sql, 3);
    confess "Entities error: zero rows retrieved to convert (highly unlikely)\n'$_'\n"
      if @rows == 0;
    return @rows;
}

1;

# Local Variables: #
# mode:cperl #
# comment-column:40 #
# cperl-comment-column:40 #
# cperl-continued-brace-offset:-2 #
# cperl-indent-level:4 #
# indent-tabs-mode:nil #
# fill-column: 78 #
# End: #
