#
# Licensed Materials - Property of IBM
#
# Rational ClearQuest
#
# (C) Copyright IBM Corp. 2005,2006  All Rights Reserved
#

package Long2clob::Subcommand;

use strict;
use Carp;
use Data::Dumper;
use Getopt::Long qw(:config require_order);

use Long2clob::Diag;
use Long2clob::PDSQL qw($COLHEAD);
use Long2clob::Operations qw(/\$OPA_/ /\$OPN_/ /\$OPS_/ /\$DBS_/ timestamp);
use Long2clob::Globals qw(/\$STM_/ /\$ERM_/ /\$OKM_/ /\%Fixed_/ $LONG2CLOB);

require Exporter;
our @ISA = qw/Exporter/;
our @EXPORT_OK=qw/%COMMANDS/;

use vars qw($VERSION);
$VERSION='1.13';

# Hash mapping subcommand names to subroutines
# callable via: $subc->{$COMMANDS{$name}}(@ARGV);
#
our %COMMANDS = ();

######################################################################

=item $subcmd = new Long2clob::Subcommand(pdsql=>$pdsql);

Return a Long2clob::Subcommand object that provides the command line API for
converting a database from LONG to CLOB.

=cut
sub new(%) {
    my ($class, %p) = @_;
    my $s = bless {%p}, ref($class) || $class;
    #
    # {pdsql} is used to talk to the database directly
    # {ops} controls atomic conversion 'operations'
    #
    confess $ERM_INTERNAL unless defined $s->{pdsql};
    $s->{ops} = new Long2clob::Operations(pdsql=>$s->{pdsql});
    return $s;
}

######################################################################

=item $db->cmd_lock()

The 'lock' method will lock a LONG database for conversion from LONG
to CLOB.

=cut
$COMMANDS{lock} = \&cmd_lock;
sub cmd_lock() {
    @_==1 or confess "$ERM_USAGE: unrecognized parameter(s): @_[1..$#_]\n";
    my $ds = new Long2clob::Diag();
    my ($s) = @_;
    #
    # Check the conversion status
    #
    my ($status, $desc) = $s->{ops}->get_conversion_status();
    die "$ERM_USAGE\n\nCurrent status: $status\n\n$desc.\n"
      unless $status eq $DBS_LONG;
    #
    # Lock the database
    #
    $s->{pdsql}->lock_for_reason($LONG2CLOB);
    #
    # Print the new status and the next subcommand to run
    #
    ($status, $desc) = $s->{ops}->get_conversion_status();
    print "\nNew status: $status\n\n$desc\n";
    exit 0;
}

######################################################################

=item $db->cmd_initialize()

The 'initialize' method plans the conversion process by analyzing the database and
loading conversion conversions into the RATL_LONGTOCLOB_OPERATIONS table.

=cut
$COMMANDS{initialize} = \&cmd_initialize;
sub cmd_initialize() {
    @_==1 or confess "$ERM_USAGE: unrecognized parameter(s): @_[1..$#_]\n";
    my $ds = new Long2clob::Diag();
    my ($s) = @_;
    #
    # Check the conversion status
    #
    my ($status, $desc) = $s->{ops}->get_conversion_status();
    die "$ERM_USAGE\n\nCurrent status: $status\n\n$desc.\n"
      unless $status eq $DBS_LOCKED;
    #
    # Initialize the database
    #
    $s->{ops}->drop_table();
    $s->{ops}->create_table();
    #
    # Fixed_SR_Info and Fixed_UserDB_Info provide partial SQL
    # to convert master and user databases from LONG to CLOB.
    #
    my $fixed = $s->{pdsql}->is_schema_repo() ? \%Fixed_SR_Info : \%Fixed_UserDB_Info;
    #
    # Plan conversion operations for all the described tables.
    #
    for my $table_name_ix (0 .. $#{$fixed->{table_name}}) {
        #
        # Get the next $table_name and its @index_names
        #
        my $table_name = $fixed->{table_name}->[$table_name_ix];
        print "$STM_ANALYZE: $table_name\n";
        my @index_names = ($fixed->{index_name}->[$table_name_ix]);
        @index_names = @{$index_names[0]} if ref($index_names[0]) eq 'ARRAY';
        #
        # Plan to rename the old LONG table and its indices.
        #
        my $index_name;
        for $index_name (@index_names) {
            $s->{ops}->plan_operation(name=>$OPN_INDEX_RENAME,
                                      args=>"${index_name},${index_name}_LONG");
        }
        $s->{ops}->plan_operation(name=>$OPN_TABLE_RENAME,
                                  args=>"${table_name},${table_name}_LONG");
        #
        # Plan to create the new CLOB table and indices.
        #
        $s->{ops}->plan_operation(name=>$OPN_TABLE_CREATE,
                                  args=>${table_name});
        for $index_name (@index_names) {
            $s->{ops}->plan_operation(name=>$OPN_INDEX_CREATE,
                                      args=>${index_name});
        }
        #
        # Plan to copy data from the old table to the new one.
        #
        $s->{ops}->plan_operation(name=>$OPN_TABLE_INSERT,
                                  args=>${table_name});
        #
        # Plan to delete the old table (indices dropped automatically).
        #
        $s->{ops}->plan_operation(name=>$OPN_TABLE_DROP,
                                  args=>"${table_name}_LONG");
    }
    #
    # Plan to convert each user entity, copying data from MULTILINE_TEXT table.
    #
    my @entities = $s->{pdsql}->getAllUserEntities()
      unless $s->{pdsql}->is_schema_repo();
    for my $entity (@entities) {
        my ($userentity_db, $fielddef_db, $fielddef_id) = map {uc} @{$entity};
        $s->{ops}->plan_operation(name=>$OPN_TABLE_ADD_COL,
                                  args=>"$userentity_db,$fielddef_db");
        $s->{ops}->plan_operation(name=>$OPN_TABLE_SET_COL,
                                  args=>"$userentity_db,$fielddef_db,$fielddef_id");
    }
    #
    # Plan to delete all MULTILINE_TEXT rows
    #
    if (! $s->{pdsql}->is_schema_repo()) {
        $s->{ops}->plan_operation(name=>$OPN_TABLE_DROP, args=>"MULTILINE_TEXT,/rows");
    }
    #
    # Record that the initialize command has finished without interruption.
    #
    $s->{ops}->plan_operation(status=>$OPS_COMPLETED,
                              name=>$OPN_INITIALIZE,
                              args=>"CONVERSION",
                              silent=>1);
    #
    # Print the new status and the next subcommand to run
    #
    ($status, $desc) = $s->{ops}->get_conversion_status();
    print "\nNew status: $status\n\n$desc\n";
    exit 0;
}

######################################################################

=item $db->cmd_status();

The 'status' method prints all the planned conversion operations.

=cut
$COMMANDS{status} = \&cmd_status;
sub cmd_status() {
    @_==1 or confess "$ERM_USAGE: unrecognized parameter(s): @_[1..$#_]\n";
    my $ds = new Long2clob::Diag();
    my ($s) = @_;
    #
    # Get the overall status of the conversion
    #
    my ($status, $desc) = $s->{ops}->get_conversion_status();
    #
    # If this status has an operations table, dump the contents.
    #

    if ($status eq $DBS_INITIALIZED || $status eq $DBS_CONVERTING ||
        $status eq $DBS_CONVERTED || $status eq $DBS_FINALIZED ||
        $status eq $DBS_UNLOCKED)
    {
        my @ops = $s->{ops}->get_operations();
        for my $op ({status=>'STATUS', name=>'OPERATION', args=>'ARGUMENTS', header=>1},
                    {status=>'------', name=>'---------', args=>'---------', header=>1},
                    @ops)
        {
            printf("%-10s %-15s %-15s\n", $op->{status}, $op->{name}, $op->{args});
            print("  ".$s->{ops}->get_sql_statement($op)."\n\n")
              if ($::opt_verbose && ! $op->{header} &&
                  $op->{name} ne $OPN_INITIALIZE && $op->{name} ne $OPN_FINALIZE);
        }
        print scalar(@ops)." operations found.\n";
    }
    print "\nCurrent status: $status\n\n$desc\n";
    exit 0;
}

######################################################################

=item $db->cmd_convert();

The 'convert' method executes planned operations from the
RATL_LONGTOCLOB_OPERATIONS table.

If the 'convert' method is interrupted, run it again and it will
resume from where it left off.

=cut
$COMMANDS{convert} = \&cmd_convert;
sub cmd_convert() {
    my $ds = new Long2clob::Diag();
    my $s = shift;
    #
    # Check the conversion status.  Note: converted is okay because user may
    # want to execute previously deferred operations.
    #
    my ($status, $desc) = $s->{ops}->get_conversion_status();
    die "$ERM_USAGE\n\nCurrent status: $status\n\n$desc.\n"
      if !($status eq $DBS_INITIALIZED or
           $status eq $DBS_CONVERTING or
           $status eq $DBS_CONVERTED);
    #
    # Start executing from the operations table, but defer some operations
    # until after the 'unlock' if -fast was given on the command line.
    #
    my $opt_fast;
    @_ = grep { ! ($_ eq '-fast' and $opt_fast = 1) } @_;
    @_==0 or confess "$ERM_USAGE: unrecognized parameter(s): @_\n";
    #
    # Get all operations and figure out what to skip, execute, or defer.
    #
    my @ops = $s->{ops}->get_operations();
    for my $op (@ops) {
        my ($status, $name, $args) = ($op->{status}, $op->{name}, $op->{args});
        #
        # The $action to take can be 'DEFER' (if it qualifies and -fast was
        # given), 'EXECUTE', or 'SKIP' (if completed in a previous run).
        #
        my $action = $OPA_EXECUTE;
        if ($status eq $OPS_COMPLETED) {
            $action = $OPA_SKIP;
        }
        elsif ($opt_fast) {
            if ($name eq $OPN_INDEX_DROP || $name eq $OPN_TABLE_DROP ||
                $name eq $OPN_TABLE_INSERT && $args eq "ATTACHMENTS_BLOB") {
                $action = $OPA_DEFER;
            }
        }
        print timestamp()." " if $::opt_verbose;
        printf("%-10s %-10s %-15s %s\n", $action, $op->{status}, $op->{name}, $op->{args});
        if ($action eq $OPA_DEFER) {
            $s->{ops}->set_operation_status($op, $OPS_DEFERRED);
        }
        elsif ($action eq $OPA_EXECUTE) {
            $s->{ops}->execute_operation($op);
        }
        elsif ($action eq $OPA_SKIP) {
        }
        else {
            confess $ERM_INTERNAL;
        }
    }
    #
    # Print the new status and the next subcommand to run
    #
    ($status, $desc) = $s->{ops}->get_conversion_status();
    print "\nNew status: $status\n\n$desc\n";
    exit 0;
}

######################################################################

=item $exit_status = $db->cmd_finalize()

Update the connection info and unlock the database.

=cut
$COMMANDS{finalize} = \&cmd_finalize;
sub cmd_finalize() {
    my $ds = new Long2clob::Diag();
    my $s = shift;

    #
    # Check the conversion status
    #
    my ($status, $desc) = $s->{ops}->get_conversion_status();
    die "$ERM_USAGE\n\nCurrent status: $status\n\n$desc.\n"
      unless $status eq $DBS_CONVERTED;

    #
    # Sanity check $status and verify this is really a CLOB database
    #
    my $scriptdef = $s->{pdsql}->is_schema_repo()
      ? "M_SCRIPTDEF" : "SCRIPTDEF";
    my $sql = <<EOF;
select count(*) clobnumdata_type$COLHEAD
  from user_tab_columns
  where table_name = '$scriptdef' and column_name = 'PROGRAM' and data_type = 'CLOB';
EOF
    my $numClobsInSanity = ($s->{pdsql}->pdsql_select($sql,1))[0]->[0];
    die "$ERM_INTERNAL\nValidation of $scriptdef table failed\nCLOB datatype not found for PROGRAM"
      unless $numClobsInSanity == 1;

    #
    # Set up values for working with the masterdb where the connect options
    # need to be updated.  No real work is needed if the master database is
    # the one being converted.
    #
    my $mastr = $s->{pdsql};            # $mastr = PDSQL connection
    my $site_name = 'MASTR';            # $site_name = key for dbglobal table

    #
    # Compute values if the converted db is a USER database.
    #
    if (! $s->{pdsql}->is_schema_repo())
    {
        my %opt;
        my $erm = <<EOF;
$ERM_MISSING_OPTION
Finalizing the USER database requires a second set of database connection
parameters after the 'finalize' subcommand which specify the MASTER schema
repository where the USER database connection options will be updated.
EOF
        local @ARGV = @_;               # parse 'finalize' parameters
        die $erm if (@ARGV<6);
        GetOptions(\%opt, 'u=s', 'p=s', 's=s','db=s', 'co=s', 'v=s')
          or die $erm;
        my @rows = $s->{pdsql}->pdsql_select("select site_name name$COLHEAD from dbglobal;",1);
        $site_name = $rows[0]->[0];
        $mastr = new Long2clob::PDSQL(%opt);
    }

    #
    # Update the connect options in the master database
    #
    my $co = $mastr->get_connect_options($site_name);
    $co =~ s/\bLOB_TYPE=LONG\b/LOB_TYPE=CLOB/ig
      or confess "$ERM_FINALIZE\nMissing LOB_TYPE=LONG in existing connect_options:\n  $co\n";
    $mastr->set_connect_options($site_name, $co);

    #
    # Record connect options have been updated.
    #
    $s->{ops}->plan_operation(status=>$OPS_COMPLETED,
                              name=>$OPN_FINALIZE,
                              args=>"LOB_TYPE=CLOB",
                              silent=>1);
    #
    # Print the new status and the next subcommand to run
    #
    ($status, $desc) = $s->{ops}->get_conversion_status();
    print "\nNew status: $status\n\n$desc\n";
    exit 0;
}

######################################################################

=item $db->cmd_unlock()

The 'unlock' method will either (a) unlock a LONG database that has
been locked for conversion to LONG but where 'convert' has NOT been
started, or (b) unlock a CLOB database that has been converted from
LONG.

=cut
$COMMANDS{unlock} = \&cmd_unlock;
sub cmd_unlock() {
    @_==1 or confess "$ERM_USAGE: unrecognized parameter(s): @_[1..$#_]\n";
    my $ds = new Long2clob::Diag();
    my ($s) = @_;
    #
    # Check the conversion status
    #
    my ($status, $desc) = $s->{ops}->get_conversion_status();
    die "$ERM_USAGE\n\nCurrent status: $status\n\n$desc.\n"
      unless $status eq $DBS_FINALIZED or $status eq $DBS_LOCKED;
    #
    # Unlock the database
    #
    $s->{pdsql}->unlock_for_reason($LONG2CLOB) or die $ERM_UNLOCK_FAILED;
    #
    # Print the new status and the next subcommand to run
    #
    ($status, $desc) = $s->{ops}->get_conversion_status();
    print "\nNew status: $status\n\n$desc\n";
    #
    # Remind the customer to convert the master database too
    #
    if (! $s->{pdsql}->is_schema_repo()) {
        print <<EOF;

Note: This user database will not be accessible by Rational ClearQuest
      version 7 clients until its master database has also been converted
      to CLOB.

EOF
    }
    #
    # Update the registry or tell the user they need to unlock the MASTR
    #
    $s->{pdsql}->update_registry()
      or print <<EOF;
Note: The connect options in your registry settings cannot be updated until
      the master database is unlocked.  If your master database is currently
      locked, the registry will be updated when you issue the unlock
      subcommand for the master database.

EOF
    exit 0;
}

######################################################################

=item $db->cmd_cleanup();

The 'cleanup' subcommand finishes deferred operations like converting
attachments and dropping unneeded tables.

=cut
$COMMANDS{cleanup} = \&cmd_cleanup;
sub cmd_cleanup() {
    @_==1 or confess "$ERM_USAGE: unrecognized parameter(s): @_[1..$#_]\n";
    my $ds = new Long2clob::Diag();
    my ($s) = @_;
    #
    # Check the conversion status
    #
    my ($status, $desc) = $s->{ops}->get_conversion_status();
    die "$ERM_USAGE\n\nCurrent status: $status\n\n$desc.\n"
      unless $status eq $DBS_UNLOCKED;
    #
    # Execute any deferred conversion operations.
    #
    my @ops = $s->{ops}->get_operations();
    my @deferred = grep { $_->{status} ne $OPS_COMPLETED } @ops;
    for my $op (@deferred) {
        print timestamp()." " if $::opt_verbose;
        printf("%-10s %-10s %-15s %s\n", $OPA_EXECUTE, $op->{status}, $op->{name}, $op->{args});
        $s->{ops}->execute_operation($op);
    }
    #
    # Drop the OPS_TABLE
    #
    $s->{ops}->drop_table();
    #
    # Print the new status and the next subcommand to run (none in this case)
    #
    ($status, $desc) = $s->{ops}->get_conversion_status();
    print "\nNew status: $status\n\n$desc\n";
    exit 0;
}

######################################################################
# The following are unsupported commands that may corrupt your database.
# Intended for testing purposes only.
######################################################################

$COMMANDS{nuke_all_tables} = \&cmd_nuke_all_tables;
sub cmd_nuke_all_tables() {
    @_==1 or confess "$ERM_USAGE: unrecognized parameter(s): @_[1..$#_]\n";
    my $ds = new Long2clob::Diag();
    my ($s) = @_;
    $s->{ops}->drop_table();
    $s->{pdsql}->pdsql("nuke_all_tables;");
    print <<EOF;
NUKE_ALL_TABLES
every database table has been dropped by force.
EOF
    exit 0;
}

$COMMANDS{uninitialize} = \&cmd_uninitialize;
sub cmd_uninitialize() {
    @_==1 or confess "$ERM_USAGE: unrecognized parameter(s): @_[1..$#_]\n";
    my $ds = new Long2clob::Diag();
    my ($s) = @_;
    $s->{ops}->drop_table();
    print <<EOF;
UNINITIALIZED
The operations table has been dropped by force without verification.
EOF
    exit 0;
}

1;

# Local Variables: #
# mode:cperl #
# comment-column:40 #
# cperl-comment-column:40 #
# cperl-continued-brace-offset:-2 #
# cperl-indent-level:4 #
# indent-tabs-mode:nil #
# fill-column: 78 #
# End: #
